@@ -1,41 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtQuick module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include "qquickloader_p_p.h"
 
@@ -45,6 +9,7 @@
 #include <private/qqmlglobal_p.h>
 
 #include <private/qqmlcomponent_p.h>
+#include <private/qqmlincubator_p.h>
 
 QT_BEGIN_NAMESPACE
 
@@ -403,7 +368,7 @@ QUrl QQuickLoader::source() const
     return d->source;
 }
 
-void QQuickLoader::setSource(const QUrl &url)
+void QQuickLoader::setSourceWithoutResolve(const QUrl &url)
 {
     setSource(url, true); // clear previous values
 }
@@ -438,9 +403,8 @@ void QQuickLoader::loadFromSource()
     }
 
     if (isComponentComplete()) {
-        QQmlComponent::CompilationMode mode = d->asynchronous ? QQmlComponent::Asynchronous : QQmlComponent::PreferSynchronous;
         if (!d->component)
-            d->component.setObject(new QQmlComponent(qmlEngine(this), d->source, mode, this), this);
+            d->createComponent();
         d->load();
     }
 }
@@ -512,6 +476,22 @@ void QQuickLoader::loadFromSourceComponent()
         d->load();
 }
 
+
+QUrl QQuickLoader::setSourceUrlHelper(const QUrl &unresolvedUrl)
+{
+    Q_D(QQuickLoader);
+
+    // 1. If setSource is called with a valid url, clear the old component and its corresponding url
+    // 2. If setSource is called with an invalid url(e.g. empty url), clear the old component but
+    // hold the url for old one.(we will compare it with new url later and may update status of loader to Loader.Null)
+    QUrl oldUrl = d->source;
+    d->clear();
+    QUrl sourceUrl = qmlEngine(this)->handle()->callingQmlContext()->resolvedUrl(unresolvedUrl);
+    if (!sourceUrl.isValid())
+        d->source = oldUrl;
+    return sourceUrl;
+}
+
 /*!
     \qmlmethod object QtQuick::Loader::setSource(url source, object properties)
 
@@ -576,25 +556,34 @@ void QQuickLoader::loadFromSourceComponent()
 
     \sa source, active
 */
-void QQuickLoader::setSource(QQmlV4Function *args)
+void QQuickLoader::setSource(const QUrl &source, QJSValue properties)
 {
-    Q_ASSERT(args);
     Q_D(QQuickLoader);
 
-    bool ipvError = false;
-    args->setReturnValue(QV4::Encode::undefined());
-    QV4::Scope scope(args->v4engine());
-    QV4::ScopedValue ipv(scope, d->extractInitialPropertyValues(args, this, &ipvError));
-    if (ipvError)
+    if (!(properties.isArray() || properties.isObject())) {
+        qmlWarning(this) << QQuickLoader::tr("setSource: value is not an object");
         return;
-
-    d->clear();
-    QUrl sourceUrl = d->resolveSourceUrl(args);
-    d->disposeInitialPropertyValues();
-    if (!ipv->isUndefined()) {
-        d->initialPropertyValues.set(args->v4engine(), ipv);
     }
-    d->qmlCallingContext.set(scope.engine, scope.engine->qmlContext());
+
+    QUrl sourceUrl = setSourceUrlHelper(source);
+
+    d->disposeInitialPropertyValues();
+    auto engine = qmlEngine(this)->handle();
+    d->initialPropertyValues.set(engine, QJSValuePrivate::takeManagedValue(&properties)->asReturnedValue());
+    d->qmlCallingContext.set(engine, engine->qmlContext());
+
+    setSource(sourceUrl, false); // already cleared and set ipv above.
+}
+
+void QQuickLoader::setSource(const QUrl &source)
+{
+    Q_D(QQuickLoader);
+
+    QUrl sourceUrl = setSourceUrlHelper(source);
+
+    d->disposeInitialPropertyValues();
+    auto engine = qmlEngine(this)->handle();
+    d->qmlCallingContext.set(engine, engine->qmlContext());
 
     setSource(sourceUrl, false); // already cleared and set ipv above.
 }
@@ -643,14 +632,15 @@ void QQuickLoaderPrivate::setInitialState(QObject *obj)
         // does, then set the item's size now before bindings are
         // evaluated, otherwise we will end up resizing the item
         // later and triggering any affected bindings/anchors.
-        if (widthValid && !QQuickItemPrivate::get(item)->widthValid)
+        if (widthValid() && !QQuickItemPrivate::get(item)->widthValid())
             item->setWidth(q->width());
-        if (heightValid && !QQuickItemPrivate::get(item)->heightValid)
+        if (heightValid() && !QQuickItemPrivate::get(item)->heightValid())
             item->setHeight(q->height());
         item->setParentItem(q);
     }
     if (obj) {
-        QQml_setParent_noEvent(itemContext, obj);
+        if (itemContext)
+            QQml_setParent_noEvent(itemContext, obj);
         QQml_setParent_noEvent(obj, q);
         itemContext = nullptr;
     }
@@ -665,7 +655,8 @@ void QQuickLoaderPrivate::setInitialState(QObject *obj)
     QV4::Scope scope(v4);
     QV4::ScopedValue ipv(scope, initialPropertyValues.value());
     QV4::Scoped<QV4::QmlContext> qmlContext(scope, qmlCallingContext.value());
-    d->initializeObjectWithInitialProperties(qmlContext, ipv, obj);
+    auto incubatorPriv = QQmlIncubatorPrivate::get(incubator);
+    d->initializeObjectWithInitialProperties(qmlContext, ipv, obj, incubatorPriv->requiredProperties());
 }
 
 void QQuickLoaderIncubator::statusChanged(Status status)
@@ -709,7 +700,6 @@ void QQuickLoaderPrivate::incubatorStateChanged(QQmlIncubator::Status status)
     emit q->progressChanged();
     if (status == QQmlIncubator::Ready)
         emit q->loaded();
-    disposeInitialPropertyValues(); // cleanup
 }
 
 void QQuickLoaderPrivate::_q_sourceLoaded()
@@ -729,15 +719,26 @@ void QQuickLoaderPrivate::_q_sourceLoaded()
         return;
     }
 
+    if (!active)
+        return;
+
     QQmlContext *creationContext = component->creationContext();
-    if (!creationContext) creationContext = qmlContext(q);
-    itemContext = new QQmlContext(creationContext);
-    itemContext->setContextObject(q);
+    if (!creationContext)
+        creationContext = qmlContext(q);
+
+    QQmlComponentPrivate *cp = QQmlComponentPrivate::get(component);
+    QQmlContext *context = [&](){
+        if (cp->isBound())
+            return creationContext;
+        itemContext = new QQmlContext(creationContext);
+        itemContext->setContextObject(q);
+        return itemContext;
+    }();
 
     delete incubator;
     incubator = new QQuickLoaderIncubator(this, asynchronous ? QQmlIncubator::Asynchronous : QQmlIncubator::AsynchronousIfNested);
 
-    component->create(*incubator, itemContext);
+    component->create(*incubator, context);
 
     if (incubator && incubator->status() == QQmlIncubator::Loading)
         updateStatus();
@@ -794,12 +795,9 @@ void QQuickLoader::componentComplete()
 {
     Q_D(QQuickLoader);
     QQuickItem::componentComplete();
-    if (active()) {
-        if (d->loadingFromSource) {
-            QQmlComponent::CompilationMode mode = d->asynchronous ? QQmlComponent::Asynchronous : QQmlComponent::PreferSynchronous;
-            if (!d->component)
-                d->component.setObject(new QQmlComponent(qmlEngine(this), d->source, mode, this), this);
-        }
+    if (active() && (status() != Ready)) {
+        if (d->loadingFromSource)
+            d->createComponent();
         d->load();
     }
 }
@@ -821,8 +819,6 @@ void QQuickLoader::itemChange(QQuickItem::ItemChange change, const QQuickItem::I
 
     This signal is emitted when the \l status becomes \c Loader.Ready, or on successful
     initial load.
-
-    The corresponding handler is \c onLoaded.
 */
 
 
@@ -937,7 +933,7 @@ void QQuickLoaderPrivate::_q_updateSize(bool loaderGeometryChanged)
 }
 
 /*!
-    \qmlproperty object QtQuick::Loader::item
+    \qmlproperty QtObject QtQuick::Loader::item
     This property holds the top-level object that is currently loaded.
 
     Since \c {QtQuick 2.0}, Loader can load any object type.
@@ -948,44 +944,13 @@ QObject *QQuickLoader::item() const
     return d->object;
 }
 
-void QQuickLoader::geometryChanged(const QRectF &newGeometry, const QRectF &oldGeometry)
+void QQuickLoader::geometryChange(const QRectF &newGeometry, const QRectF &oldGeometry)
 {
     Q_D(QQuickLoader);
     if (newGeometry != oldGeometry) {
         d->_q_updateSize();
     }
-    QQuickItem::geometryChanged(newGeometry, oldGeometry);
-}
-
-QUrl QQuickLoaderPrivate::resolveSourceUrl(QQmlV4Function *args)
-{
-    QV4::Scope scope(args->v4engine());
-    QV4::ScopedValue v(scope, (*args)[0]);
-    QString arg = v->toQString();
-    if (arg.isEmpty())
-        return QUrl();
-
-    QQmlContextData *context = scope.engine->callingQmlContext();
-    Q_ASSERT(context);
-    return context->resolvedUrl(QUrl(arg));
-}
-
-QV4::ReturnedValue QQuickLoaderPrivate::extractInitialPropertyValues(QQmlV4Function *args, QObject *loader, bool *error)
-{
-    QV4::Scope scope(args->v4engine());
-    QV4::ScopedValue valuemap(scope, QV4::Encode::undefined());
-    if (args->length() >= 2) {
-        QV4::ScopedValue v(scope, (*args)[1]);
-        if (!v->isObject() || v->as<QV4::ArrayObject>()) {
-            *error = true;
-            qmlWarning(loader) << QQuickLoader::tr("setSource: value is not an object");
-        } else {
-            *error = false;
-            valuemap = v;
-        }
-    }
-
-    return valuemap->asReturnedValue();
+    QQuickItem::geometryChange(newGeometry, oldGeometry);
 }
 
 QQuickLoader::Status QQuickLoaderPrivate::computeStatus() const
@@ -1033,6 +998,23 @@ void QQuickLoaderPrivate::updateStatus()
     }
 }
 
-#include <moc_qquickloader_p.cpp>
+void QQuickLoaderPrivate::createComponent()
+{
+    Q_Q(QQuickLoader);
+    const QQmlComponent::CompilationMode mode = asynchronous
+            ? QQmlComponent::Asynchronous
+            : QQmlComponent::PreferSynchronous;
+    if (QQmlContext *context = qmlContext(q)) {
+        if (QQmlEngine *engine = context->engine()) {
+            component.setObject(new QQmlComponent(
+                                    engine, context->resolvedUrl(source), mode, q), q);
+            return;
+        }
+    }
+
+    qmlWarning(q) << "createComponent: Cannot find a QML engine.";
+}
 
 QT_END_NAMESPACE
+
+#include <moc_qquickloader_p.cpp>
